# coding: utf-8

import unittest
from unittest.mock import Mock
import pytest
import pandas
from pandas.testing import assert_series_equal, assert_frame_equal
from snakemake.io import Wildcards


class TestUnitUtils(unittest.TestCase):
    def setUp(self):
        self.units = pandas.read_table(
            "tests/utils/files/units.tsv",
            dtype=str
        ).set_index(["sample", "type", "flowcell", "lane", "barcode"], drop=False).sort_index()

        self.units_2 = pandas.read_table(
            "tests/utils/files/units_2.tsv",
            dtype=str
        ).set_index(["sample", "type"], drop=False).sort_index()
        self.sample_NA12878 = {
            "L1": Wildcards(fromdict={
                'sample': 'NA12878', "flowcell": "HKTG2BGXG", "type": "N", "lane": "L001", "barcode": "ACGGAACA+ACGAGAAC"
            }),
            "L2": Wildcards(fromdict={
                'sample': 'NA12878', "flowcell": "HKTG2BGXG", "type": "N", "lane": "L002", "barcode": "CCGGAACA+ACGAGAAC"
            })
        }
        self.sample_NA13878 = Wildcards(fromdict={
            'sample': 'NA13878', "flowcell": "HKTG2BGXG", "type": "N", "lane": "L001", "barcode": "GCGGAACA+ACGAGAAC"
        })
        self.sample_NA22878 = Wildcards(fromdict={
            'sample': 'NA22878', "flowcell": "HKTG2BGXG", "type": "N", "lane": "L001", "barcode": "TGGGGGGG+ACGAGAAC"
        })
        self.sample_NA12978 = Wildcards(fromdict={
            'sample': 'NA12978', "flowcell": "HLCF3DRXY", "type": "N", "lane": "L001", "barcode": "AAGGAACA+ACGAGAAC"
        })
        self.sample_BE12878 = {
            "N_L1": Wildcards(fromdict={
                'sample': 'BE12878', "flowcell": "HLCF3DRXY", "type": "N", "lane": "L001", "barcode": "ACGGAACA+ACGAGAAC"
            }),
            "N_L2": Wildcards(fromdict={
                'sample': 'BE12878', "flowcell": "HLCF3DRXY", "type": "N", "lane": "L002", "barcode": "AGGGAACA+ACGAGAAC"
            }),
            "T_L3": Wildcards(fromdict={
                'sample': 'BE12878', "flowcell": "HLCF3DRXY", "type": "T", "lane": "L003", "barcode": "ATGGAACA+ACGAGAAC"
            }),
            "R_L4": Wildcards(fromdict={
                'sample': 'BE12878', "flowcell": "HLCF3DRXY", "type": "R", "lane": "L004", "barcode": "ACAGAACA+ACGAGAAC"
            })
        }

    def tearDown(self):
        pass

    def test_get_unit(self):
        from hydra_genetics.utils.units import get_unit
        assert_series_equal(
            get_unit(self.units, self.sample_NA12878["L1"]),
            self.units.loc[('NA12878', "N", "HKTG2BGXG", "L001", "ACGGAACA+ACGAGAAC")].dropna()
        )
        assert_series_equal(
            get_unit(self.units, self.sample_NA12878["L2"]),
            self.units.loc[('NA12878', "N", "HKTG2BGXG", "L002", "CCGGAACA+ACGAGAAC")].dropna()
        )
        assert_series_equal(
            get_unit(self.units, self.sample_NA13878),
            self.units.loc[('NA13878', "N", "HKTG2BGXG", "L001", "GCGGAACA+ACGAGAAC")].dropna()
        )
        assert_series_equal(
            get_unit(self.units, self.sample_NA22878),
            self.units.loc[('NA22878', "N", "HKTG2BGXG", "L001", "TGGGGGGG+ACGAGAAC")].dropna()
        )
        assert_series_equal(
            get_unit(self.units, self.sample_NA12978),
            self.units.loc[('NA12978', "N", "HLCF3DRXY", "L001", "AAGGAACA+ACGAGAAC")].dropna()
        )
        assert_series_equal(
            get_unit(self.units, self.sample_BE12878["N_L1"]),
            self.units.loc[('BE12878', "N", "HLCF3DRXY", "L001", "ACGGAACA+ACGAGAAC")].dropna()
        )
        assert_series_equal(
            get_unit(self.units, self.sample_BE12878["N_L2"]),
            self.units.loc[('BE12878', "N", "HLCF3DRXY", "L002", "AGGGAACA+ACGAGAAC")].dropna()
        )
        assert_series_equal(
            get_unit(self.units, self.sample_BE12878["T_L3"]),
            self.units.loc[('BE12878', "T", "HLCF3DRXY", "L003", "ATGGAACA+ACGAGAAC")].dropna()
        )
        assert_series_equal(
            get_unit(self.units, self.sample_BE12878["R_L4"]),
            self.units.loc[('BE12878', "R", "HLCF3DRXY", "L004", "ACAGAACA+ACGAGAAC")].dropna()
        )

    def test_get_fastq_file(self):
        from hydra_genetics.utils.units import get_fastq_file
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA12878["L1"]),
            self.units.loc[('NA12878', "N", "HKTG2BGXG", "L001", "ACGGAACA+ACGAGAAC")].dropna()['fastq1']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA12878["L2"]),
            self.units.loc[('NA12878', "N", "HKTG2BGXG", "L002", "CCGGAACA+ACGAGAAC")].dropna()['fastq1']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA13878),
            self.units.loc[('NA13878', "N", "HKTG2BGXG", "L001", "GCGGAACA+ACGAGAAC")].dropna()['fastq1']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA22878),
            self.units.loc[('NA22878', "N", "HKTG2BGXG", "L001", "TGGGGGGG+ACGAGAAC")].dropna()['fastq1']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA12978),
            self.units.loc[('NA12978', "N", "HLCF3DRXY", "L001", "AAGGAACA+ACGAGAAC")].dropna()['fastq1']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_BE12878["N_L1"]),
            self.units.loc[('BE12878', "N", "HLCF3DRXY", "L001", "ACGGAACA+ACGAGAAC")].dropna()['fastq1']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_BE12878["N_L2"]),
            self.units.loc[('BE12878', "N", "HLCF3DRXY", "L002", "AGGGAACA+ACGAGAAC")].dropna()['fastq1']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_BE12878["T_L3"]),
            self.units.loc[('BE12878', "T", "HLCF3DRXY", "L003", "ATGGAACA+ACGAGAAC")].dropna()['fastq1']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_BE12878["R_L4"]),
            self.units.loc[('BE12878', "R", "HLCF3DRXY", "L004", "ACAGAACA+ACGAGAAC")].dropna()['fastq1']
        )

        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA12878["L1"], "fastq2"),
            self.units.loc[('NA12878', "N", "HKTG2BGXG", "L001", "ACGGAACA+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA12878["L2"], "fastq2"),
            self.units.loc[('NA12878', "N", "HKTG2BGXG", "L002", "CCGGAACA+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA13878, "fastq2"),
            self.units.loc[('NA13878', "N", "HKTG2BGXG", "L001", "GCGGAACA+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA22878, "fastq2"),
            self.units.loc[('NA22878', "N", "HKTG2BGXG", "L001", "TGGGGGGG+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_NA12978, "fastq2"),
            self.units.loc[('NA12978', "N", "HLCF3DRXY", "L001", "AAGGAACA+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_BE12878["N_L1"], "fastq2"),
            self.units.loc[('BE12878', "N", "HLCF3DRXY", "L001", "ACGGAACA+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_BE12878["N_L2"], "fastq2"),
            self.units.loc[('BE12878', "N", "HLCF3DRXY", "L002", "AGGGAACA+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_BE12878["T_L3"], "fastq2"),
            self.units.loc[('BE12878', "T", "HLCF3DRXY", "L003", "ATGGAACA+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertEqual(
            get_fastq_file(self.units, self.sample_BE12878["R_L4"], "fastq2"),
            self.units.loc[('BE12878', "R", "HLCF3DRXY", "L004", "ACAGAACA+ACGAGAAC")].dropna()['fastq2']
        )
        self.assertNotEqual(
            get_fastq_file(self.units, self.sample_NA12878["L1"], "fastq2"),
            self.units.loc[('NA12878', "N", "HKTG2BGXG", "L001", "ACGGAACA+ACGAGAAC")].dropna()['fastq1']
        )

        with self.assertRaisesRegex(ValueError, "Incorrect input value error fastq3: expected fastq1 or fastq2"):
            self.assertTrue(get_fastq_file(self.units, self.sample_NA12878["L1"], "fastq3"))

        with self.assertRaises(KeyError):
            self.assertTrue(
                get_fastq_file(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                               "flowcell": "HLCF3DRXY",
                                                               "type": "N",
                                                               "lane": "L005",
                                                               "barcode": "AAAAAA+CCCCCC"}))
            )

    def test_get_fastq_adapter(self):
        from hydra_genetics.utils.units import get_fastq_adapter
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_NA12878["L1"]),
            "ACGT,TGCA1"
        )
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_NA12878["L2"]),
            "ACGT,TGCA2"
        )
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_NA13878),
            "ACGT,TGCA3"
        )
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_NA22878),
            "ACGT,TGCA4"
        )
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_NA12978),
            "ACGT,TGCA5"
        )
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_BE12878["N_L1"]),
            "ACGT,TGCA6"
        )
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_BE12878["N_L2"]),
            "ACGT,TGCA7"
        )
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_BE12878["T_L3"]),
            "ACGT,TGCA8"
        )
        self.assertEqual(
            get_fastq_adapter(self.units, self.sample_BE12878["R_L4"]),
            "ACGT,TGCA9"
        )

        with self.assertRaises(KeyError):
            self.assertTrue(
                get_fastq_adapter(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                                  "flowcell": "HLCF3DRXY",
                                                                  "type": "N",
                                                                  "lane": "L005",
                                                                  "barcode": "AAAA+CCCC"}))
            )

    def test_get_units(self):
        from hydra_genetics.utils.units import get_units
        self.assertEqual(
            len(get_units(self.units, Wildcards(fromdict={'sample': 'NA12878', "type": "N"}))),
            2
        )
        self.assertEqual(
            len(get_units(self.units, Wildcards(fromdict={'sample': 'NA13878', "type": "N"}))),
            1
        )
        self.assertEqual(
            get_units(self.units, Wildcards(fromdict={'sample': 'NA13878', "type": "N"}))[0],
            list(self.units.loc[('NA13878', "N")].dropna().itertuples())[0]
        )
        self.assertEqual(
            len(get_units(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "N"}))),
            2
        )
        self.assertEqual(
            len(get_units(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "T"}))),
            1
        )
        self.assertEqual(
            len(get_units(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "R"}))),
            1
        )
        self.assertEqual(
            len(get_units(self.units_2, Wildcards(fromdict={'sample': 'NA12878', "type": "N"}))),
            2
        )

        # small dataframe that produces a pandas.Series in get_units
        units3 = pandas.DataFrame(
            dict(
                sample=["NA12878", "NA12878", "NA13878"],
                type=["N", "T", "N"],
            )
        ).set_index(["sample", "type"], drop=False)
        self.assertEqual(len(get_units(units3, Wildcards(fromdict={"sample": "NA12878", "type": "N"}))), 1)

    def test_get_unit_barcodes(self):
        from hydra_genetics.utils.units import get_unit_barcodes
        self.assertEqual(
            get_unit_barcodes(self.units, Wildcards(fromdict={'sample': 'NA12878',
                                                              'flowcell': 'HKTG2BGXG',
                                                              'type': "N"})),
            {'ACGGAACA+ACGAGAAC', 'CCGGAACA+ACGAGAAC'}
        )
        self.assertEqual(
            get_unit_barcodes(self.units, Wildcards(fromdict={'sample': 'NA13878',
                                                              'flowcell': 'HKTG2BGXG',
                                                              'type': "N"})),
            {'GCGGAACA+ACGAGAAC'}
        )
        self.assertEqual(
            get_unit_barcodes(self.units, Wildcards(fromdict={'sample': 'NA22878',
                                                              'flowcell': 'HKTG2BGXG',
                                                              'type': "N"})),
            {'TGGGGGGG+ACGAGAAC'}
        )
        self.assertEqual(
            get_unit_barcodes(self.units, Wildcards(fromdict={'sample': 'NA12978',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "N"})),
            {'AAGGAACA+ACGAGAAC'}
        )
        self.assertEqual(
            get_unit_barcodes(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "N"})),
            {'ACGGAACA+ACGAGAAC', 'AGGGAACA+ACGAGAAC'}
        )
        self.assertEqual(
            get_unit_barcodes(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "T"})),
            {'ATGGAACA+ACGAGAAC'}
        )
        self.assertEqual(
            get_unit_barcodes(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "R"})),
            {'ACAGAACA+ACGAGAAC'}
        )

    def test_get_fastq_files(self):
        from hydra_genetics.utils.units import get_fastq_files
        self.assertEqual(
            len(get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA12878', "type": "N", "read": "fastq1"}))),
            2
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA12878', "type": "N", "read": "fastq1"})),
            ['input/1_R1_L1.fq.gz', 'input/2_R1_L2.fq.gz']
        )
        self.assertEqual(
            len(get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA12878', "type": "N", "read": "fastq2"}))),
            2
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA12878', "type": "N", "read": "fastq2"})),
            ['input/1_R2_L1.fq.gz', 'input/2_R2_L2.fq.gz']
        )
        self.assertEqual(
            len(get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA13878', "type": "N", "read": "fastq1"}))),
            1
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA13878', "type": "N", "read": "fastq1"})),
            ['input/3_R1_L1.fq.gz']
        )
        self.assertEqual(
            len(get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA13878', "type": "N", "read": "fastq2"}))),
            1
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA13878', "type": "N", "read": "fastq1"})),
            ['input/3_R1_L1.fq.gz']
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'NA13878', "type": "N", "read": "fastq2"})),
            ['input/3_R2_L1.fq.gz']
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "N", "read": "fastq1"})),
            ['input/6_R1_L1.fq.gz', 'input/7_R1_L2.fq.gz']
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "N", "read": "fastq2"})),
            ['input/6_R2_L1.fq.gz', 'input/7_R2_L2.fq.gz']
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "T", "read": "fastq1"})),
            ['input/8_R1_L3.fq.gz']
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "T", "read": "fastq2"})),
            ['input/8_R2_L3.fq.gz']
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "R", "read": "fastq1"})),
            ['input/9_R1_L4.fq.gz']
        )
        self.assertEqual(
            get_fastq_files(self.units, Wildcards(fromdict={'sample': 'BE12878', "type": "R", "read": "fastq2"})),
            ['input/9_R2_L4.fq.gz']
        )

    def test_get_unit_machine(self):
        from hydra_genetics.utils.units import get_unit_machine
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'NA12878',
                                                             'flowcell': 'HKTG2BGXG',
                                                             'type': "N",
                                                             'lane': "L001",
                                                             'barcode': "ACGGAACA+ACGAGAAC"})),
            'NDX550220'
        )
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'NA12878',
                                                             'flowcell': 'HKTG2BGXG',
                                                             'type': "N",
                                                             'lane': "L002",
                                                             'barcode': "CCGGAACA+ACGAGAAC"})),
            'NDX550220'
        )
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'NA13878',
                                                             'flowcell': 'HKTG2BGXG',
                                                             'type': "N",
                                                             'lane': "L001",
                                                             'barcode': "GCGGAACA+ACGAGAAC"})),
            'NDX550220'
        )
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'NA22878',
                                                             'flowcell': 'HKTG2BGXG',
                                                             'type': "N",
                                                             'lane': "L001",
                                                             'barcode': "TGGGGGGG+ACGAGAAC"})),
            'NDX550220'
        )
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'NA12978',
                                                             'flowcell': 'HLCF3DRXY',
                                                             'type': "N",
                                                             'lane': 'L001',
                                                             'barcode': "AAGGAACA+ACGAGAAC"})),
            'M03273'
        )
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                             'flowcell': 'HLCF3DRXY',
                                                             'type': "N",
                                                             'lane': 'L001',
                                                             'barcode': "ACGGAACA+ACGAGAAC"})),
            'A00687'
        )
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                             'flowcell': 'HLCF3DRXY',
                                                             'type': "N",
                                                             'lane': 'L002',
                                                             'barcode': "AGGGAACA+ACGAGAAC"})),
            'A00687'
        )
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                             'flowcell': 'HLCF3DRXY',
                                                             'type': "T",
                                                             'lane': 'L003',
                                                             'barcode': "ATGGAACA+ACGAGAAC"})),
            'A00687'
        )
        self.assertEqual(
            get_unit_machine(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                             'flowcell': 'HLCF3DRXY',
                                                             'type': "R",
                                                             'lane': 'L004',
                                                             'barcode': "ACAGAACA+ACGAGAAC"})),
            'A00687'
        )

    def test_get_unit_platform(self):
        from hydra_genetics.utils.units import get_unit_platform
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'NA12878',
                                                              'flowcell': 'HKTG2BGXG',
                                                              'type': "N",
                                                              'lane': "L001",
                                                              'barcode': "ACGGAACA+ACGAGAAC"})),
            'nextseq'
        )
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'NA12878',
                                                              'flowcell': 'HKTG2BGXG',
                                                              'type': "N",
                                                              'lane': "L002",
                                                              'barcode': "CCGGAACA+ACGAGAAC"})),
            'nextseq'
        )
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'NA13878',
                                                              'flowcell': 'HKTG2BGXG',
                                                              'type': "N",
                                                              'lane': "L001",
                                                              'barcode': "GCGGAACA+ACGAGAAC"})),
            'nextseq'
        )
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'NA22878',
                                                              'flowcell': 'HKTG2BGXG',
                                                              'type': "N",
                                                              'lane': "L001",
                                                              'barcode': "TGGGGGGG+ACGAGAAC"})),
            'nextseq'
        )
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'NA12978',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "N",
                                                              'lane': 'L001',
                                                              'barcode': "AAGGAACA+ACGAGAAC"})),
            'miniseq'
        )
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "N",
                                                              'lane': 'L001',
                                                              'barcode': "ACGGAACA+ACGAGAAC"})),
            'novaseq'
        )
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "N",
                                                              'lane': 'L002',
                                                              'barcode': "AGGGAACA+ACGAGAAC"})),
            'novaseq'
        )
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "T",
                                                              'lane': 'L003',
                                                              'barcode': "ATGGAACA+ACGAGAAC"})),
            'novaseq'
        )
        self.assertEqual(
            get_unit_platform(self.units, Wildcards(fromdict={'sample': 'BE12878',
                                                              'flowcell': 'HLCF3DRXY',
                                                              'type': "R",
                                                              'lane': 'L004',
                                                              'barcode': "ACAGAACA+ACGAGAAC"})),
            'novaseq'
        )

    def test_get_platforms(self):
        from hydra_genetics.utils.units import get_platforms
        self.assertEqual(
            get_platforms(self.units, Wildcards(fromdict={'sample': 'NA12878', 'type': "N"})),
            {'nextseq'}
        )
        self.assertEqual(
            get_platforms(self.units, Wildcards(fromdict={'sample': 'NA22878', 'type': "N"})),
            {'nextseq'}
        )
        self.assertEqual(
            get_platforms(self.units, Wildcards(fromdict={'sample': 'NA12978', 'type': "N"})),
            {'miniseq'}
        )
        self.assertEqual(
            get_platforms(self.units, Wildcards(fromdict={'sample': 'BE12878', 'type': "N"})),
            {'novaseq'}
        )

    def test_get_unit_types(self):
        from hydra_genetics.utils.units import get_unit_types
        self.assertEqual(
            get_unit_types(self.units, Wildcards(fromdict={'sample': 'NA12878'})),
            {'N'}
        )
        self.assertEqual(
            get_unit_types(self.units, Wildcards(fromdict={'sample': 'BE12878'})),
            {'N', 'T', 'R'}
        )

    def test_get_units_per_flowcell(self):
        from hydra_genetics.utils.units import get_units_per_flowcell
        self.assertEqual(
            len(get_units_per_flowcell(self.units, Wildcards(fromdict={"flowcell": "HKTG2BGXG"}))),
            3
        )
        self.assertEqual(
            len(get_units_per_flowcell(self.units, Wildcards(fromdict={"flowcell": "HLCF3DRXY"}))),
            4
        )



# coding: utf-8

import unittest
import yaml


class TestResourcesUtils(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_merge(self):
        from hydra_genetics.utils.misc import merge
        dict1a = {'bwa_mem': {'extra': "some settings"}}
        dict1b = {'bwa_mem': {'extra': "some settings"}}
        dict2 = {'bwa_mem': {'cpu': 20, 'time': '12:00:00'}}
        dict1a.update(dict2)
        self.assertEqual({'bwa_mem': {'cpu': 20, 'time': '12:00:00'}}, dict1a)
        dict1b = merge(dict1b, dict2)
        self.assertEqual({'bwa_mem': {'extra': "some settings", 'cpu': 20, 'time': '12:00:00'}}, dict1b)

    def test_extract_chr(self):
        from hydra_genetics.utils.misc import extract_chr
        self.assertEqual(extract_chr("tests/utils/files/ref.fasta.fai"), [
                                     'chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 'chr9', 'chr10', 'chr11',
                                     'chr12', 'chr13', 'chr14', 'chr15', 'chr16', 'chr17', 'chr18', 'chr19', 'chr20', 'chr21',
                                     'chr22', 'chrX', 'chrY'])
        self.assertEqual(extract_chr("tests/utils/files/ref.fasta.fai", ['chr6', 'chr17', 'chrX']), [
                                     'chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr7', 'chr8', 'chr9', 'chr10', 'chr11', 'chr12',
                                     'chr13', 'chr14', 'chr15', 'chr16', 'chr18', 'chr19', 'chr20', 'chr21', 'chr22', 'chrY',
                                     'chrM'])
        self.assertEqual(extract_chr(""), [''])

    def test_variable_replacement(self):
        with open("tests/utils/files/config_variable_replacement.yaml") as f:
            config = yaml.load(f, Loader=yaml.loader.SafeLoader)
        print(config)
        self.assertEqual(config,
                         {'PROJECT': 'MY variable', 'bwa_mem': {'extra': 'some settings {{PROJECT}}'}})

        from hydra_genetics.utils.misc import replace_dict_variables
        self.assertEqual(replace_dict_variables(config),
                         {'PROJECT': 'MY variable', 'bwa_mem': {'extra': 'some settings MY variable'}})



# coding: utf-8

import pandas
import snakemake
import warnings


def get_unit(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards) -> pandas.Series:
    """
    function used to extract one unit(row) from units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, type, flowcell, and lane
    Returns:
        Series containing data of the selected row
    Raises:
        raises an exception (KeyError) if no unit can be extracted from the Dataframe
    """
    unit = units.loc[(wildcards.sample, wildcards.type, wildcards.flowcell, wildcards.lane, wildcards.barcode)]
    return unit


def get_fastq_file(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards, read_pair: str = "fastq1") -> str:
    """
    function used to extract path for one unit(row) from units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, type, flowcell, and lane
        read_pair: fast1 or fastq2
    Returns:
        path for fastq file as a str
    Raises:
        raises an exception (KeyError) if no unit can be extracted from the Dataframe
    """
    unit = get_unit(units, wildcards)
    if read_pair not in ["fastq1", "fastq2"]:
        raise ValueError("Incorrect input value error {}: expected {} or {}".format(read_pair, "fastq1", "fastq2"))
    return unit[read_pair]


def get_fastq_adapter(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards) -> str:
    """
    function used to extract adapters for one unit(row) from units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, type, flowcell, and lane
    Returns:
        return adapter with format seq1,seq2
    Raises:
        raises an exception (KeyError) if no unit can be extracted from the Dataframe
    """
    unit = get_unit(units, wildcards)
    return unit["adapter"]


def get_unit_barcodes(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards) -> str:
    """
    function used to extract barcode for one unit(row) from units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, type, and flowcell
    Returns:
        return barcodes
    Raises:
        raises an exception (KeyError) if no unit can be extracted from the Dataframe
    """
    return set([u.barcode for u in units.loc[
        (units['sample'] == wildcards.sample) &
        (units['flowcell'] == wildcards.flowcell) &
        (units['type'] == wildcards.type),
    ].itertuples()])


def get_unit_machine(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards) -> str:
    """
    function used to extract machine for one unit(row) from units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, type, flowcell, and lane
    Returns:
        return machine id
    Raises:
        raises an exception (KeyError) if no unit can be extracted from the Dataframe
    """
    unit = get_unit(units, wildcards)
    return unit["machine"]


def get_unit_platform(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards) -> str:
    """
    function used to extract platform for one unit(row) from units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, type, flowcell, and lane
    Returns:
        return platform
    Raises:
        raises an exception (KeyError) if no unit can be extracted from the Dataframe
    """
    unit = get_unit(units, wildcards)
    return unit["platform"]


def get_unit_flowcell(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards) -> str:
    """
    function used to extract flowcell for one unit(row) from units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, type, flowcell, and lane
    Returns:
        return flowcell
    Raises:
        raises an exception (KeyError) if no unit can be extracted from the Dataframe
    """
    unit = get_unit(units, wildcards)
    return unit["flowcell"]


def get_units(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards, type: str = None) -> pandas.DataFrame:
    """
    function used to extract one or more units from units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample and type (optional, can also be passed as an argument).
        type: N,T or R
    Returns:
        all units from the DataFrame that can be filtereted out using sample name
        and unit type (N,T,R)
    Raises:
        raises an exception (KeyError) if no unit(s) can be extracted from the Dataframe
    """
    if type is None:
        files = units.loc[(wildcards.sample, wildcards.type)]
    else:
        files = units.loc[(wildcards.sample, type)]
    if isinstance(files, pandas.Series):
        files = pandas.DataFrame(
            [[f[1] for f in files.items()], ], columns=[f[0] for f in files.items()]
        ).set_index(units.index.names)
    return [file for file in files.itertuples()]


def get_fastq_files(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards, type: str = None):
    """
    function used to extract all fastq files for a sample with a sepecific type
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, read and type (optional, can also be passed as an argument).
        type: N,T or R
    Returns:
        all units from the DataFrame that can be filtereted out using sample name
        ,fastq (1 or 2) and type type (N,T,R)
    Raises:
        raises an exception (KeyError) if no unit(s) can be extracted from the Dataframe
    """
    return [getattr(file, wildcards.read) for file in get_units(units, wildcards, type)]


def get_platforms(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards) -> set:
    """
    function used to extract all uniq platform values for a sample and type combination found in units.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, type.
    Returns:
        set of string
    Raises:
        raises an exception (KeyError) if no unit(s) can be extracted from the Dataframe
    """
    return set([u.platform for u in units.loc[(wildcards.sample, wildcards.type)].itertuples()])


def get_unit_types(units: pandas.DataFrame, sample: str) -> set:
    """
    function used to extract all types of units found for a sample in units.tsv (N,T,R)
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample.
    Returns:
        set of type types ex set("N","T")
    Raises:
        raises an exception (KeyError) if no unit(s) can be extracted from the Dataframe
    """
    return set([u.type for u in units.loc[(sample,)].itertuples()])


def get_units_per_flowcell(units: pandas.DataFrame, wildcards: snakemake.io.Wildcards):
    """
    function used to extract all sample and type combinations for one sequencing flowcell
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample, flowcell.
    Returns:
        tuple with sample and type
    Raises:
        raises an exception (KeyError) if no unit(s) can be extracted from the Dataframe
    """
    return set([(u.sample, u.type) for u in units[units.flowcell == wildcards.flowcell].itertuples()])



# coding: utf-8

import typing
import pandas
import snakemake


def get_sample(samples: pandas.DataFrame, wildcards: snakemake.io.Wildcards) -> pandas.Series:
    """
    function used to extract one sample(row) from sample.tsv
    Args:
        samples: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/samples.schema.tsv
        wildcards: wildcards object with at least the following wildcard names
               sample
    Returns:
        Series containing data of the selected row
    Raises:
        raises an exception (KeyError) if no sample can be extracted from the Dataframe
    """
    return samples.loc[(wildcards.sample)].dropna()


def get_samples(samples: pandas.DataFrame) -> typing.List[str]:
    """
    function used to extract all sample found in samples.tsv
    Args:
        units: DataFrame generate by importing a file following schema defintion
               found in pre-alignment/workflow/schemas/units.schema.tsv
    Returns:
        List of strings with all sample names
    """
    return [sample.Index for sample in samples.itertuples()]



# coding: utf-8

from collections.abc import Mapping
from copy import deepcopy
import os
import re
from snakemake.sourcecache import GithubFile, LocalGitFile


def get_module_snakefile(config, repo, path, tag):
    ''' return a complete snakefile path, either pointing to github or a local repo'''
    if 'hydra_local_path' in config:
        repo_path = os.path.join(config['hydra_local_path'], repo)
        if not os.path.exists(repo_path):
            ' Remove possible organization part from repo path '
            repo_path_v2 = os.path.join(config['hydra_local_path'], os.sep.join(repo.split(os.sep)[1:]))
            if not os.path.exists(repo_path_v2):
                raise FileNotFoundError(f"Repo folder {repo_path} or {repo_path_v2} doesn't exist!")
            else:
                repo_path = repo_path_v2
        if not os.path.isdir(repo_path):
            raise FileNotFoundError(f"{repo_path} isn't a folder!")
        return LocalGitFile(repo_path, path=path, tag=tag)
    else:
        return GithubFile(repo, path=path, tag=tag)


def merge(dict1, dict2):
    ''' Return a new dictionary by merging two dictionaries recursively. '''

    result = deepcopy(dict1)
    for key, value in dict2.items():
        if isinstance(value, Mapping):
            result[key] = merge(result.get(key, {}), value)
        else:
            result[key] = deepcopy(dict2[key])
    return result


def extract_chr(file, filter_out=["chrM"]):
    chr = None
    if os.path.exists(file):
        with open(file) as lines:
            chr = [line.split("\t")[0] for line in lines]
        return [c for c in chr if c not in filter_out]
    return [""]


def replace_dict_variables(config):
    '''
        Make it possible to defina a variable in a  vaue string using ´{{}}´,
        ex {{PROJECT_VARIABLE}}. PROJECT_VARIABLE must exist as a variable
        in the yaml file and will be used to replace all occurences of
        {{PROJECT_VARAIBLE}} in the yaml file.
    '''
    def update_dict(temp_config):
        for k in temp_config:
            if type(temp_config[k]) is dict:
                temp_config[k] = update_dict(temp_config[k])
            elif type(temp_config[k]) is str:
                match = re.findall(r'\{\{([A-Za-z0-9_]+)\}\}', temp_config[k])
                for m in match:
                    temp_config[k] = temp_config[k].replace("{{" + m + "}}", config[m])
        return temp_config
    return update_dict(config)




import click
import logging
import os
import re
import sys

import hydra_genetics.utils
from hydra_genetics.commands.prep_pipeline_env import environment
from hydra_genetics.commands.references import references
from hydra_genetics.commands.create import PipelineCreate, RuleCreate, CreateInputFiles
import rich.console
import rich.logging
import rich.traceback


# Set up logging as the root logger
# Submodules should all traverse back to this
log = logging.getLogger()


def rich_force_colors():
    """
    Check if any environment variables are set to force Rich to use coloured output
    """
    if os.getenv("GITHUB_ACTIONS") or os.getenv("FORCE_COLOR") or os.getenv("PY_COLORS"):
        return True
    return None


def run():
    # Set up rich stderr console
    stderr = rich.console.Console(stderr=True, force_terminal=rich_force_colors())

    # Set up the rich traceback
    rich.traceback.install(console=stderr, width=200, word_wrap=True, extra_lines=1)

    stderr.print("\u001b[36m" + r'  _  _  _  _  ____  ____   __         ___  ____  __ _  ____  ____  __  ___  ____ ',
                 highlight=False)
    stderr.print("\u001b[36m" + r' / )( \( \/ )(    \(  _ \ / _\  ___  / __)(  __)(  ( \(  __)(_  _)(  )/ __)/ ___) ',
                 highlight=False)
    stderr.print("\u001b[36m" + r' ) __ ( )  /  ) D ( )   //    \(___)( (_ \ ) _) /    / ) _)   )(   )(( (__ \___ \ ',
                 highlight=False)
    stderr.print("\u001b[36m" + r' \_)(_/(__/  (____/(__\_)\_/\_/      \___/(____)\_)__)(____) (__) (__)\___)(____/ ',
                 highlight=False)
    stderr.print("\u001b[36m hydra-core/tools version {}\033[0m\n\n".format(hydra_genetics.__version__), highlight=False)

    cli()


@click.group(help="CLI tool to prepare and initialize snakemake projects")
@click.option("--verbose", is_flag=True, default=False, help="Print verbose output to the console.")
@click.option("-l", "--log-file", help="Save a verbose log to a file.", metavar="<filename>")
@click.option('--debug/--no-debug', default=False)
def cli(verbose, log_file, debug):
    loglevel = logging.WARNING
    if debug:
        loglevel = logging.DEBUG
    elif verbose:
        loglevel = logging.INFO

    logging.basicConfig(level=loglevel)
    log.setLevel(loglevel)
    log.handlers.clear()
    log.addHandler(
        rich.logging.RichHandler(
            level=loglevel,
            console=rich.console.Console(stderr=True, force_terminal=rich_force_colors()),
            show_time=False,
            markup=True,
        )
    )
    log.propagate = False

    if log_file:
        log_fh = logging.FileHandler(log_file, encoding="utf-8")
        log_fh.setLevel(logging.DEBUG)
        log_fh.setFormatter(logging.Formatter("[%(asctime)s] %(name)-20s [%(levelname)-7s]  %(message)s"))
        log.addHandler(log_fh)


def validate_wf_name_prompt(ctx, opts, value):
    """Force the workflow name to meet the hydra-core requirements"""
    if not re.match(r"^[a-z][a-z0-9_]+$", value):
        click.echo("Invalid workflow name: must be lowercase without punctuation.")
        value = click.prompt(opts.prompt)
        return validate_wf_name_prompt(ctx, opts, value)
    return value


def validate_rule_name_prompt(ctx, opts, value):
    """Force the rule name to meet the hydra-core requirements"""
    if value is not None and not re.match(r"^[a-z0-9_]+$", value):
        click.echo("Invalid command/tool formatting: only lowercase and '_' is allowed.")
        value = click.prompt(opts.prompt)
        return validate_rule_name_prompt(ctx, opts, value)
    return value


@cli.command(short_help="create bare bone project, pipeline or module")
@click.option(
    "-n",
    "--name",
    prompt="Workflow Name",
    required=True,
    callback=validate_wf_name_prompt,
    type=str,
    help="The name of your new pipeline",
)
@click.option("-d", "--description", prompt=True, required=True, type=str, help="A short description of your pipeline")
@click.option("-a", "--author", prompt=True, required=True, type=str, help="Name of the main author(s)")
@click.option("-e", "--email", prompt=True, required=True, type=str, help="E-mail(s) of the main author(s)")
@click.option("--version", type=str, default="0.0.1", help="The initial version number to use")
@click.option("--min-snakemake-version", type=str, default="6.8.0", help="Min snakemake version")
@click.option("-g", "--git-user", prompt=True, required=True, help="User name of main git user(s)")
@click.option("--no-git", is_flag=True, default=False, help="Do not create git repo")
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite output directory if it already exists")
@click.option("-o", "--outdir", type=str, help="Output directory for new pipeline (default: pipeline name)")
def create_pipeline(name, description, author, email, version, min_snakemake_version, git_user, no_git, force, outdir):
    pipeline = PipelineCreate(name, description, author, email, version, min_snakemake_version, git_user, no_git, force, outdir)
    pipeline.init_pipeline()


@cli.command(short_help="add rule to project")
@click.option(
    "-c",
    "--command",
    prompt="command name/rule name",
    required=True,
    callback=validate_rule_name_prompt,
    type=str,
    help="command that will be run, will be used to name the rule",
)
@click.option(
    "-t",
    "--tool",
    required=False,
    callback=validate_rule_name_prompt,
    type=str,
    default=None,
    help="tool that will be used to run the command, if provided it will be used during the naming of the "
         "rule, ex samtools, optional",
)
@click.option(
    "-m",
    "--module",
    prompt="name of module/workflow",
    required=True,
    callback=validate_wf_name_prompt,
    type=str,
    help="name module/pipeline where rule will be added. Expected folder structure is module_name_or_pipeline/workflow/, "
         " the rule will be added to a subfolder named rules, env.yaml to a subfolder named envs.")
@click.option(
        "-a",
        "--author",
        prompt=True,
        required=True,
        type=str,
        help="Name of the main author(s)")
@click.option(
        "-e",
        "--email",
        prompt=True,
        required=True,
        type=str,
        help="E-mail(s) of the main author(s)")
@click.option(
        "-o",
        "--outdir",
        type=str,
        help="Output directory for where module is located (default: current dir)")
def create_rule(command, tool, module, author, email, outdir):
    rule = RuleCreate(command, module, author, email, tool, outdir)
    rule.init_rule()


@cli.command(short_help="create input-files, samples.tsv and units.tsv")
@click.option(
    "-d",
    "--directory",
    multiple=True,
    prompt="directory/directories",
    required=True,
    type=str,
    help="path to dir where fastq-files should be looked for.")
@click.option(
        "-o",
        "--outdir",
        type=str,
        help="Output directory for where rule will be added (default: current dir)")
@click.option(
        "-p",
        "--platform",
        type=str,
        help="Sequence platform that the data originate from, ex nextseq, miseq. Default Illumina",
        default="Illumina")
@click.option(
        "-t",
        "--sample-type",
        type=str,
        help="Sample type, N|T|R, default T",
        default="T")
@click.option(
        "-s",
        "--sample-regex",
        type=str,
        help="Regex used find fastq files and to extract samplefrom filename, default '([A-Za-z0-9-]+)_.+.gz$'",
        default=r"([A-Za-z0-9-]+)_.+\.fastq.gz")
@click.option(
        "-n",
        "--read-number-regex",
        type=str,
        help="Regex used to extract read number from filename (note only number value), default '_R([1-2]{1})_001'",
        default="_R([12]{1})_")
@click.option(
        "-a",
        "--adapters",
        type=str,
        help="adapter sequence, comma sepearated",
        default="AGATCGGAAGAGCACACGTCTGAACTCCAGTCA,AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGT")
@click.option(
        "--post-file-modifier",
        type=str,
        help="add string to output files",
        default=None)
@click.option(
        "-f",
        "--force",
        help="overwrite existing files",
        is_flag=True)
@click.option(
        "-b",
        "--default-barcode",
        help="default barcode value that should be used when the fastq files are missing barcode information in their header, "
             "if not set the tool will fail if barcode can not be extracted",
        type=str,
        default=None)
@click.option(
        "--tc",
        help="tumor contet",
        type=float,
        default=1.0)
@click.option(
        "--validate",
        help="see if fastq contain multipl runs/lanes by comparing first and last "
             "read. Note will take time since whole file need to be parsed.",
        is_flag=True)
@click.option(
        "--ask",
        help="ask user input when inconsistent machine id or flow cell id are found, only asked when --validate is set.",
        is_flag=True)
@click.option(
        "--th",
        help="if occurences of a concesuns base in barcode is below this value a warning will be printed",
        type=float,
        default=0.9)
@click.option(
        "--nreads",
        help="number of reads that will be used to generate consensus barcode.",
        type=int,
        default=200)
@click.option(
        "--every",
        help="select every N reads for validation.",
        type=int,
        default=1000)
def create_input_files(directory, outdir, post_file_modifier, platform, sample_type,
                       sample_regex, read_number_regex, adapters, tc, force, default_barcode, validate, ask, th, nreads, every):
    input_files = CreateInputFiles(directory, outdir, post_file_modifier, platform, sample_type,
                                   sample_regex, read_number_regex, adapters, tc, force, default_barcode, validate, ask, th,
                                   nreads, every)
    input_files.init()


cli.add_command(references)

cli.add_command(environment)

if __name__ == "__main__":
    run()



# coding: utf-8

import click
import hashlib
import json
import logging
import os
import pathlib
import re
import requests
import yaml

from hydra_genetics.utils.misc import merge
from hydra_genetics.utils.io.reference import fetch_reference_data, validate_reference_data


@click.group("references", short_help="setup/validate reference files")
def references():
    pass


@references.command(short_help="make sure that all reference file exist and are valid")
@click.option(
    "-c",
    "--config-file",
    prompt="path",
    required=True,
    multiple=True,
    type=str,
    help="config file used by pipeline",
)
@click.option(
    "-v",
    "--validation-file",
    prompt="path",
    required=True,
    multiple=True,
    type=str,
    help="json/yaml file defining required reference files",
)
@click.option(
    "-p",
    "--path-to-ref-data",
    required=False,
    type=str,
    default="",
    help="path to where reference files are stored",
)
@click.option(
    "-s",
    "--skip-regex",
    required=False,
    multiple=True,
    default=[r"config\/\S+\.(json|yaml|yml|tsv|html|hg19)"],
    type=str,
    help="regex for warning skip",
)
def validate(config_file, validation_file, path_to_ref_data, skip_regex):
    def is_file_or_folder(possible_file):
        '''
            Function used to make sure that the provided argument is a file.

            Parameters:
                possible_file (string): path for file

            Returns:
                bool: if file has been found
        '''
        if not isinstance(possible_file, str):
            return False

        if re.match("^[0-9.]+$", possible_file):
            return False

        extension = pathlib.Path(possible_file)
        # docker container aren't files
        if ":" in possible_file or possible_file.endswith(".sif"):
            return False
        # Skip files that are configured per analysis/site
        if possible_file.endswith("samples.tsv") \
           or possible_file.endswith("resources.yaml") \
           or possible_file.endswith("units.tsv") \
           or possible_file.endswith("output_list.yaml") \
           or possible_file.endswith("output_list.json") \
           or possible_file.endswith("snakemake-wrappers"):
            logging.debug(f"Ignore: {possible_file}")
            return False
        # If a extension can be found we consider it as a file
        if extension.suffix:
            # File
            return True
        else:
            if re.search(r"\/([A-Za-z0-9_.-]*)", possible_file):
                # Folder
                return True
            return False

    def locate_possible_files_in_config(config_data, files=[], skip_regex=[]):
        '''
            function used to extract files from provided config

            Parameters:
                config_data (dict): loaded configuration
                file_list (list): list used to store found files

            Returns:
                files (list): all files found in the config dict
        '''
        for config in config_data:
            possible_file = config_data[config]
            if isinstance(possible_file, dict):
                files = locate_possible_files_in_config(possible_file, files, skip_regex)
            else:
                if not isinstance(possible_file, str):
                    continue
                result = re.search(r"(\S+)", possible_file)
                if result:
                    for f in re.findall(r"(\S+)", possible_file):
                        skip_file = False
                        for skip in skip_regex:
                            if re.match(skip, f):
                                skip_file = True
                                break

                        if not skip_file and is_file_or_folder(f):
                            files.append(f)
        return files

    # Load and merge multiple config
    config_data = {}
    for config in config_file:
        # Note that duplicate entries will be replaced with latest data load
        config_data = merge(config_data, yaml.safe_load(open(config)))

    validation_data = {}
    # Load and merge multiple validation files
    for validation in validation_file:
        # Note that duplicate entries will be replaced with latest data load
        if validation.endswith(".json"):
            validation_data = merge(validation_data, json.load(open(validation)))
        elif validation.endswith(".yaml") or validation.endswith(".yml"):
            validation_data = merge(validation_data, yaml.safe_load(open(validation)))
        else:
            raise IOError(f"Unknown file format passed as validation data: {validation}")

    # All files that we should validate, note that all may not be specified in the validation files
    possible_files = [*set(locate_possible_files_in_config(config_data, skip_regex=skip_regex))]

    # List of files that haven't been validated and counters for success and failures
    possible_files, files_not_in_config, found, counter_pass, counter_fail = validate_reference_data(validation_data,
                                                                                                     path_to_ref_data,
                                                                                                     possible_files)

    logging.info(f"Files/Folders pass: {counter_pass}, fail: {counter_fail}")
    if len(possible_files) > 0:
        logging.warning(f"Found more possible files in config ({', '.join(possible_files)}) that haven't been validated!")

    if len(files_not_in_config) > 0:
        logging.warning("The following files in provided validation files couldn't"
                        f" be located in the provided config: ({', '.join(files_not_in_config)})")

    if counter_fail > 0:
        print(f"PASS: {counter_pass}, FAILED: {counter_fail}")
        exit(1)
    else:
        print(f"PASS: {counter_pass}")


@references.command(short_help="download reference data, if needed")
@click.option(
    "-v",
    "--validation-file",
    prompt="path",
    required=True,
    multiple=True,
    type=str,
    help="json/yaml file defining required reference files",
)
@click.option(
    "-o",
    "--output-dir",
    required=False,
    type=str,
    default="",
    help="path to where reference files will be stored",
)
@click.option(
    "-f",
    "--force",
    required=False,
    type=bool,
    is_flag=True,
    default=False,
    help="replace files even though they exist (skip checksum validation)",
)
def download(validation_file, output_dir, force):
    validation_data = {}
    # Load and merge multiple validation files
    for validation in validation_file:
        # Note that duplicate entries will be replaced with latest data load
        if validation.endswith(".json"):
            validation_data = merge(validation_data, json.load(open(validation)))
        elif validation.endswith(".yaml") or validation.endswith(".yml"):
            validation_data = merge(validation_data, yaml.safe_load(open(validation)))
        else:
            raise IOError(f"Unknown file format passed as validation data: {validation}")

    # Make sure that output directory exists
    if output_dir and not os.path.exists(output_dir):
        os.mkdir(output_dir)

    fetched_list, failed_list, skipped_list, files_fetched, files_failed, files_skipped = fetch_reference_data(validation_data,
                                                                                                               output_dir,
                                                                                                               force=force)

    if files_failed > 0:
        logging.error(f"Failed to retrieve {files_failed}")
        logging.error(f"Failed: {', '.join(failed_list)}")

    if files_skipped > 0:
        logging.info(f"Skipped {files_skipped} since checksum didn't change")
        logging.debug(f"Skipped: {', '.join(skipped_list)}")

    if files_fetched > 0:
        logging.info(f"Retrieved {files_fetched} files")
        logging.debug(f"Retrieved: {', '.join(fetched_list)}")

    print(f"UPDATED: {files_fetched}")
    print(f"NOT UPDATED: {files_skipped}")

    return 0



# vim: syntax=python tabstop=4 expandtab
# coding: utf-8

import logging
import os
import shutil
import tempfile
import unittest

logger = logging.getLogger(__name__).addHandler(logging.NullHandler())


class TestHotspots(unittest.TestCase):
    def setUp(self):
        # create fixtures
        self.tempdir = tempfile.mkdtemp()
        with open(os.path.join(self.tempdir, "hotspot"), 'w') as hotspots:
            hotspots.write("#Chr\tStart\tEnd\tGene\tCDS_mutation_syntax\tAA_mutation_syntax\tReport\tcomment\tExon\tAccession_number\n")  # noqa
            hotspots.write("NC_000001.10\t115252190\t115252349\tNRAS\t-\t-\tindel\t-\texon4\tNM_002524\n")
            hotspots.write("NC_000001.10\t115252202\t115252202\tNRAS\tc.C438\tp.A146\thotspot\t-\texon4\tNM_002524\n")
            hotspots.write("NC_000017.10\t37880979\t37881164\tERBB2\t-\t-\tindel\t-\texon20\tNM_004448\n")
            hotspots.write("NC_000012.11\t25398279\t25398279\tKRAS\tc.G40\tp.V14\tregion\t-\texon2\tNM_004985\n")
            hotspots.write("NC_000017.10\t37880986\t37880987\tERBB2\t-\tp.Y772\tregion_all\t-\texon20\tNM_004448\n")

    def tearDown(self):
        # delete fixtures
        shutil.rmtree(self.tempdir)

    def test_header_parsing(self):
        from hydra_genetics.utils.io.hotspot import Reader
        reader = Reader(os.path.join(self.tempdir, "hotspot"))
        self.assertEqual(10, len(reader.header.keys()))
        self.assertEqual(0, reader.header["CHR"])
        self.assertEqual(1, reader.header["START"])
        self.assertEqual(2, reader.header["END"])
        self.assertEqual(3, reader.header["GENE"])
        self.assertEqual(4, reader.header["CDS_MUTATION_SYNTAX"])
        self.assertEqual(5, reader.header["AA_MUTATION_SYNTAX"])
        self.assertEqual(6, reader.header["REPORT"])
        self.assertEqual(7, reader.header["COMMENT"])
        self.assertEqual(8, reader.header["EXON"])
        self.assertEqual(9, reader.header["ACCESSION_NUMBER"])

    def test_record_parsing(self):
        from hydra_genetics.utils.io.hotspot import Reader
        from hydra_genetics.utils.models.hotspot import ReportClass
        reader = Reader(os.path.join(self.tempdir, "hotspot"))

        record = reader.next()
        self.assertEqual("NC_000001.10", record.CHROMOSOME)
        self.assertEqual(115252190, record.START)
        self.assertEqual(115252349, record.END)
        self.assertEqual("NRAS", record.GENE)
        self.assertEqual("-", record.CDS_MUTATION_SYNTAX)
        self.assertEqual("-", record.AA_MUTATION_SYNTAX)
        self.assertEqual(ReportClass.indel, record.REPORT)
        self.assertEqual("-", record.COMMENT)
        self.assertEqual("exon4", record.EXON)
        self.assertEqual("NM_002524", record.ACCESSION_NUMBER)
        self.assertEqual(160, len(record.VARIANTS))
        self.assertListEqual([{'extended': False, "variants": []} for i in range(160)], record.VARIANTS)

        record = reader.next()
        self.assertEqual("NC_000001.10", record.CHROMOSOME)
        self.assertEqual(115252202, record.START)
        self.assertEqual(115252202, record.END)
        self.assertEqual("NRAS", record.GENE)
        self.assertEqual("c.C438", record.CDS_MUTATION_SYNTAX)
        self.assertEqual("p.A146", record.AA_MUTATION_SYNTAX)
        self.assertEqual(ReportClass.hotspot, record.REPORT)
        self.assertEqual("-", record.COMMENT)
        self.assertEqual("exon4", record.EXON)
        self.assertEqual("NM_002524", record.ACCESSION_NUMBER)
        self.assertEqual(1, len(record.VARIANTS))
        self.assertEqual([{'extended': False, "variants": []}], record.VARIANTS)

        record = reader.next()
        self.assertEqual("NC_000017.10", record.CHROMOSOME)
        self.assertEqual(37880979, record.START)
        self.assertEqual(37881164, record.END)
        self.assertEqual("ERBB2", record.GENE)
        self.assertEqual("-", record.CDS_MUTATION_SYNTAX)
        self.assertEqual("-", record.AA_MUTATION_SYNTAX)
        self.assertEqual(ReportClass.indel, record.REPORT)
        self.assertEqual("-", record.COMMENT)
        self.assertEqual("exon20", record.EXON)
        self.assertEqual("NM_004448", record.ACCESSION_NUMBER)
        self.assertEqual(186, len(record.VARIANTS))
        self.assertEqual([{'extended': False, "variants": []} for i in range(186)], record.VARIANTS)

        record = reader.next()
        self.assertEqual("NC_000012.11", record.CHROMOSOME)
        self.assertEqual(25398279, record.START)
        self.assertEqual(25398279, record.END)
        self.assertEqual("KRAS", record.GENE)
        self.assertEqual("c.G40", record.CDS_MUTATION_SYNTAX)
        self.assertEqual("p.V14", record.AA_MUTATION_SYNTAX)
        self.assertEqual(ReportClass.region, record.REPORT)
        self.assertEqual("-", record.COMMENT)
        self.assertEqual("exon2", record.EXON)
        self.assertEqual("NM_004985", record.ACCESSION_NUMBER)
        self.assertEqual(1, len(record.VARIANTS))
        self.assertEqual([{'extended': False, "variants": []}], record.VARIANTS)

        record = reader.next()
        self.assertEqual("NC_000017.10", record.CHROMOSOME)
        self.assertEqual(37880986, record.START)
        self.assertEqual(37880987, record.END)
        self.assertEqual("ERBB2", record.GENE)
        self.assertEqual("-", record.CDS_MUTATION_SYNTAX)
        self.assertEqual("p.Y772", record.AA_MUTATION_SYNTAX)
        self.assertEqual(ReportClass.region_all, record.REPORT)
        self.assertEqual("-", record.COMMENT)
        self.assertEqual("exon20", record.EXON)
        self.assertEqual("NM_004448", record.ACCESSION_NUMBER)
        self.assertEqual(2, len(record.VARIANTS))
        self.assertEqual([{'extended': False, "variants": []}, {'extended': False, "variants": []}], record.VARIANTS)

        with self.assertRaises(StopIteration):
            reader.next()


if __name__ == '__main__':
    import logging
    import sys
    logging.basicConfig(level=logging.CRITICAL, stream=sys.stdout, format='%(message)s')
    unittest.main()



import gzip
import hashlib
import logging
import os
import requests
import shutil
import tarfile
import tempfile

# Expected input file format
# --------------------------
# rule_name1:
#     variable1:
#        entry
#
# rule_name2:
#     variable2:
#        entry
#
#
# --------------------------
#
# Expected formats for an entry
# A file
# {checksum: md5sum_value, type: file, path: path_to_file, url: url_path}
# Compressed file that will be extracted
# {checksum: md5sum_value, compressed_checksum: md5sum_value, type: file, path: path_storage_location, url: url_path}
# Folder (it must be compressed)
# {type: folder, compressed_checksum: md5sum_value, path: folder_storage_location,
#   content_checksum: { file1:md5sum_value, file2:md5sum_value}, url: url_path}
# content_checksum is optional, when it's not set the only validation made is the existence of the folder


def fetch_reference_data(validation_data, output_dir,
                         fetched=[], failed=[], skipped=[],
                         files_fetched=0, files_failed=0, files_skipped=0,
                         force=False):
    '''
        Function used to validate entries found in validation files

        Parameters:
            validation_data (dict):
            output_dir (str): path to where data will be stored
            fetched (list): list with files that have been retrieved.
            failed (list): list with files that couldn't correctly be retrieved
            skipped (list): list with files that weren't retrieved, checksum haven't changed
            files_fetched (int): counter variable for files that have been retrieved.
            files_failed (int): counter variable for files that couldn't correctly be retrieved
            files_skipped (int): counter variable for files that weren't retrieved, checksum haven't changed

        Returns:
            (list, list, list, int, int, int): (fetched files, failed files, skipped files,
                                                counter fetched files, counter failed files , counter skipped files)
    '''
    for k, value in validation_data.items():
        # If validation entry
        if "path" in value:
            content_path = os.path.join(output_dir, value['path'])
            content_type = value['type']

            # Some entries may not be fetchable and only added for validation
            if 'url' not in value:
                logging.debug(f"Content {content_path} contains no url entry, i.e will not be retrieved!")
                continue

            # Check if content needs to be updated
            update_needed = update_needed_for_entry(value, output_dir)

            if update_needed or force:
                logging.info(f"updating {content_path}")
                with tempfile.TemporaryDirectory() as tmpdirname:
                    temp_content_holder = os.path.join(tmpdirname, "tempfile")

                    # Fetch content and merge any split files
                    fetch_url_content(value['url'], temp_content_holder, tmpdirname)

                    checksum_value = value['compressed_checksum'] if 'compressed_checksum' in value else value['checksum']

                    if not checksum_validate_file(temp_content_holder, checksum_value, content_path):
                        files_failed += 1
                        failed.append(content_path)
                    else:
                        if 'folder' in content_type and os.path.isdir(content_path):
                            shutil.rmtree(content_path)
                        if 'compressed_checksum' in value:
                            if extract_compressed_data(value, content_type, content_path, temp_content_holder, force):
                                logging.info(f"folder {content_path} retrieved")
                                files_fetched += 1
                            else:
                                failed.append(content_path)
                                files_failed += 1
                        else:
                            logging.info(f"file {content_path} retrieved")
                            files_fetched += 1
                            move_content(temp_content_holder, content_path)
            else:
                logging.debug(f"skipped {content_path}")
                files_skipped += 1
                skipped += content_path
        else:
            # Nested entry, recursively process content
            fetched, failed, skipped, files_fetched, files_failed, files_skipped = fetch_reference_data(validation_data[k],
                                                                                                        output_dir,
                                                                                                        fetched,
                                                                                                        failed,
                                                                                                        skipped,
                                                                                                        files_fetched,
                                                                                                        files_failed,
                                                                                                        files_skipped,
                                                                                                        force)
    return fetched, failed, skipped, files_fetched, files_failed, files_skipped


def fetch_url_content(url, content_holder, tmpdir) -> None:
    """
        Fetch content from the provided url and make sure that the downloaded file
        has the correct md5 value.

        Parameters:
            url (string): url to file
            content_holder (string): path where the data will be saved
            tmpdir: folder where we can save data temporary


    """
    if isinstance(url, dict):
        counter = 1
        list_of_temp_files = []
        for part_url, part_checksum in url.items():
            temp_file = os.path.join(tmpdir, f"file{counter}")
            list_of_temp_files.append(temp_file)
            r = requests.get(part_url, allow_redirects=True)
            open(temp_file, 'wb').write(r.content)
            if not checksum_validate_file(temp_file, part_checksum):
                logging.info(f"Failed to retrieved part {counter}: {part_url}, expected {calculated_md5}, got {part_checksum}")
                return False
            else:
                logging.debug(f"Retrieved part {counter}: {part_url}")
            counter += 1
        with open(content_holder, 'wb') as writer:
            logging.debug(f"Merge {list_of_temp_files} into {content_holder}")
            for temp_content in list_of_temp_files:
                with open(temp_content, 'rb') as reader:
                    for line in reader:
                        writer.write(line)
    else:
        r = requests.get(url, allow_redirects=True)
        open(content_holder, 'wb').write(r.content)


def validate_reference_data(validation_data, path_to_ref_data,
                            file_list=[], not_found_in_config=[], found=[],
                            counter_pass=0, counter_fail=0):
    '''
        Validate all entries in the validation dict

        Parameters:
            validation_data (dict): load json/yaml file in dict format
                                    ex {key: {key_s1: {path: value}, key_s2: {path: value}, k2: {path: value}}
            path_to_ref_data (str): path to where data is stored
            file_list (list): list with files in config file
            not_found_in_config (list): files in loaded validation that couldn't be found in the configuration file
            counter_pass (int): counter variable for files that passed test
            counter_fail (int): counter variable for files that did not pass test

        Returns:
            (list, int, int): list with files that haven't been validated, pass counter, fail counter
    '''
    def track_files(path, file_list=[], not_found_in_config=[], found=[]):
        if path in found:
            pass
        else:
            file_found = ""
            for f in file_list:
                if f.endswith(path) or f == path:
                    file_found = f
                    break
            if file_found:
                file_list.remove(f)
                found.append(path)
            else:
                if path not in found:
                    not_found_in_config.append(path)
        return file_list, not_found_in_config, found

    for k, item in validation_data.items():
        # A validation entry
        if "path" in item:
            file_path = os.path.join(path_to_ref_data, item['path'])
            if not os.path.exists(file_path):
                counter_fail += 1
                logging.error(f"Could not locate: {file_path}")
                file_list, not_found_in_config, found = track_files(item['path'], file_list, not_found_in_config, found)
            else:
                if 'checksum' not in item and "content_checksum" not in item:
                    counter_pass += 1
                    logging.debug(f"{item['path']} found, no checksum validation!")
                elif "content_checksum" in item:
                    logging.debug(f"Folder found: {item['path']}")
                    _, failed, not_found = checksum_validate_content(item['content_checksum'], file_path)
                    if not failed and not not_found:
                        counter_pass += 1
                    else:
                        counter_fail += 1
                else:
                    if checksum_validate_file(file_path, validation_data[k]['checksum']):
                        counter_pass += 1
                    else:
                        counter_fail += 1
                file_list, not_found_in_config, found = track_files(item['path'], file_list, not_found_in_config, found)
        else:
            # Nested entry, recursively process content
            (file_list, not_found_in_config, found,
             counter_pass, counter_fail) = validate_reference_data(validation_data[k],
                                                                   path_to_ref_data,
                                                                   file_list, not_found_in_config, found,
                                                                   counter_pass, counter_fail)
    return file_list, not_found_in_config, found, counter_pass, counter_fail


def update_needed_for_entry(item, parent_dir="./"):
    """
        Checks if entry needs to be update, i.e checksum has change or the file/folder doesn't exist

        Patameters:
            item (dict): see top part of this page
            parent_dir (string): path to where content will be saved

        Returns:
            bool: if the entry needs to be updated
    """
    if item['path'].startswith("/"):
        content_path = item['path']
    else:
        content_path = os.path.join(parent_dir, item['path'])

    if 'file' in item['type'] and not os.path.isfile(content_path) or \
       'folder' in item['type'] and not os.path.isdir(content_path):
        logging.debug(f"{content_path} not found")
        return True
    else:
        if "checksum" in item:
            return not checksum_validate_file(content_path, item['checksum'])
        elif "content_checksum" in item:
            _, failed, not_found = checksum_validate_content(item['content_checksum'], content_path)
            if failed or not_found:
                return True
            else:
                return False
        else:
            logging.debug(f"{content_path} not validation done")


def checksum_validate_content(file_checksums, parent_dir=None, print_path_name=None) -> (int, int, int):
    """
        Used to validate content of a folders using a dict with path:md5sum values.

        Parameters:
            file_checksum (dict): dict with the following format {file_path: md5sum_value, }
            parent_dir (string): path to where the file is located
            print_path_name (string): when storing a file/folder using a temporary name this makes it possible to
                                      print a more human readable string.

        Returns:
            (int, int, int): file that passed checksum validation, files that failed checksum validation, files not found
    """
    passed = 0
    failed = 0
    not_found = 0
    failed_files = []
    files_not_found = []
    for file, checksum in file_checksums.items():
        full_file_path = os.path.join(parent_dir if parent_dir is not None else "./", file)
        if os.path.isfile(full_file_path):
            if checksum_validate_file(
                os.path.join(full_file_path),
                checksum,
                file
            ):
                passed += 1
            else:
                failed_files.append(file)
                failed += 1
        else:
            logging.debug(f"{full_file_path} not found")
            files_not_found.append(file)
            not_found += 1
    logging.debug(f"folder content validation, valid: {passed}, invalid {failed}, missing {not_found}")
    logging.debug(f"failed files: {failed_files}")
    logging.debug(f"failed files: {files_not_found}")
    return passed, failed, not_found


def checksum_validate_file(file, expected_checksum, print_path_name=None) -> bool:
    """
        function used to validate checksum.

        Parameters:
            file (str): path to file
            expected_checksum (str): md5sum value
            print_path_name (str): when storing a file/folder using a temporary name this makes it possible to
                                   print a more human readable string.

        Returns
            bool: true passed validation
    """
    calculated_md5 = hashlib.md5(open(file, 'rb').read()).hexdigest()
    print_name = print_path_name if print_path_name is not None else file
    if calculated_md5 == expected_checksum:
        logging.debug(f"{print_name}: valid checksum")
        return True
    else:
        logging.error(f"{print_name}: invalid checksum, got {calculated_md5}, expected {expected_checksum}")
        return False


def extract_compressed_data(item, content_type, content_path, temp_content_holder, force=False) -> bool:
    """
        Function used to extract compressed content, both files and folders.

        Parameters:
            item (dict): a dict with checksum or content_checksum (a dict with path:md5sum values)
            content_path (string): path to where the file or folder will be saved
            temp_content_folder (string): path to where content can be temporary stored

        Returns:
            bool: if extraction was succefull
    """
    extracted_content_path = f"{temp_content_holder}_extracted"
    if "file" in content_type:
        with gzip.open(temp_content_holder, 'rb') as file:
            with open(extracted_content_path, 'wb') as writer:
                for line in file:
                    writer.write(line)
        if checksum_validate_file(extracted_content_path, item['checksum']):
            move_content(extracted_content_path, content_path)
            return True
        else:
            return False
    else:
        with tarfile.open(temp_content_holder, mode="r|gz") as tar:
            tar.extractall(extracted_content_path)
        if 'content_checksum' in item:
            if checksum_validate_content(item['content_checksum'], extracted_content_path, content_path):
                if force and os.path.isdir(content_path):
                    shutil.rmtree(content_path)
                move_content(extracted_content_path, content_path)
                return True
            else:
                return False
        else:
            move_content(extracted_content_path, content_path)
            return True


def move_content(old_path, new_path):
    """
        Helper function that will make sure that all
        parent folders for new path exists.

        Parameters:
            old_path (string): location of data that will be moved
            new_path (string): where the data will be saved
    """
    parent_dir = os.path.dirname(new_path)
    if not os.path.isdir(parent_dir):
        os.makedirs(parent_dir)
    shutil.move(old_path, new_path)



# coding: utf-8

import yaml


def load_resources(config, file: str = "resources.yaml") -> dict:
    """
        helper function used to load resources.yaml file and update config
        with the information without replacing the current content.
    """
    from hydra_genetics.utils.misc import merge
    with open(file) as file:
        return merge(config, yaml.load(file, Loader=yaml.FullLoader))



import pysam
from pysam import VariantFile
import re
import statistics
import warnings
import logging

from hydra_genetics.utils.models.hotspot import ReportClass


def add_contigs_to_header(input_name, output_name, contig_file, assembly):
    from src.lib.data.files.reference import InfoImporter
    info = InfoImporter(contig_file)
    input_vcf = VariantFile(input_name, 'r')
    for key in info:
        input_vcf.header.contigs.add(key, length=info[key]['length'], assembly=assembly)
    output_vcf = VariantFile(output_name, 'w', header=input_vcf.header)
    for record in input_vcf:
        output_vcf.write(record)


def format_report_type(hotspot):
    if hotspot.REPORT == ReportClass.region or hotspot.REPORT == ReportClass.region_all:
        return "3-check"
    elif hotspot.REPORT == ReportClass.hotspot:
        return "1-hotspot"
    else:
        raise Exception("Unhandled case")


def get_annotation_data(data_extracter, mapper):
    return lambda variant, field: data_extracter(variant, mapper[field])


def get_annotation_data_vep(field_dict):
    def extractor(variant, info_name):
        data = variant.info['CSQ'][0].split("|")[field_dict[info_name]]
        if len(data) == 0:
            return None
        return data
    return extractor


def get_depth(gvcf_file, sample, chr, start, stop, depth_flag='DP'):
    depth = [r.samples[sample][depth_flag] for r in gvcf_file.fetch(chr, start, stop)]
    if len(depth) > 1:
        return statistics.mean(depth)
    elif len(depth) == 1:
        return depth[0]
    else:
        return 0


def regex_extract(value, regex):
    search_result = re.search(regex, value)
    if search_result is None:
        return "-"
    else:
        return search_result[0]


def get_info_field(variant, info_name):
    if isinstance(variant, pysam.VariantRecord):
        try:
            data = variant.info[info_name]
            if isinstance(data, tuple):
                return ",".join(map(str, data))
            return data
        except KeyError:
            return None
    return None


def get_annotation_data_format(sample_index):
    return lambda variant, field: variant.samples[sample_index].get(field, None)


def get_annotation_data_info(variant, info_name):
    return variant.info.get(info_name, None)


def get_report_type(variant, hotspot):
    if isinstance(variant, pysam.VariantRecord):
        if hotspot.REPORT == ReportClass.region or hotspot.REPORT == ReportClass.region_all:
            return "3-check"
        if is_indel(variant) or is_multibp_sub(variant):
            return "2-indel"
    return "1-hotspot"


def get_sample_value(variant, sample):
    if isinstance(variant, pysam.VariantRecord):
        return variant.samples[sample]


def get_read_level(read_levels, rd):
    try:
        for (level, depth_status, analyzable) in read_levels:
            rd = int(rd)
            if rd >= int(level):
                return (depth_status, analyzable)
    except ValueError:
        pass
    return ("-", "zero")


def is_indel(variant):
    if len(variant.alleles) != 2:
        raise Exception("Unhandled case: " + str(variant.alleles))
    return (len(variant.alleles[0]) == 1 and len(variant.alleles[1]) > 1) or \
           (len(variant.alleles[1]) == 1 and len(variant.alleles[0]) > 1)


def is_multibp_sub(variant):
    if len(variant.alleles) != 2:
        raise Exception("Unhandled case: " + str(variant.alleles))
    return len(variant.alleles[0]) > 1 and len(variant.alleles[1]) > 1


def clinical_flagged(variant):
    if isinstance(variant, pysam.VariantRecord):
        rs = get_annotation_data_info(variant, 'snp138').startswith("rs")
        nonflagged = get_annotation_data_info(variant, 'snp138NonFlagged')
        if rs and ("-" == nonflagged or nonflagged == "."):
            return "Yes"
        else:
            return "No"


def get_vaf(variant, sample):
    if isinstance(variant, pysam.VariantRecord):
        depth, ref, var = get_depth(variant, sample)
        return int(var)/(int(depth))



# coding: utf-8

import unittest
import pandas as pandas
from snakemake.io import Wildcards


class TestSampleUtils(unittest.TestCase):
    def setUp(self):
        self.samples = pandas.read_table("tests/utils/files/samples.tsv", dtype=str).set_index("sample", drop=False)

    def tearDown(self):
        pass

    def test_get_sample(self):
        from hydra_genetics.utils.samples import get_sample
        sample = get_sample(self.samples, Wildcards(fromdict={'sample': 'NA12878'}))
        self.assertEqual(get_sample(self.samples, Wildcards(fromdict={'sample': 'NA12878'}))['tumor_content'], '0.1')
        self.assertEqual(get_sample(self.samples, Wildcards(fromdict={'sample': 'NA13878'}))['tumor_content'], '0.2')
        self.assertEqual(get_sample(self.samples, Wildcards(fromdict={'sample': 'NA22878'}))['tumor_content'], '0.3')
        self.assertEqual(get_sample(self.samples, Wildcards(fromdict={'sample': 'NA12978'}))['tumor_content'], '0.4')
        self.assertEqual(get_sample(self.samples, Wildcards(fromdict={'sample': 'BE12878'}))['tumor_content'], '0.5')

    def test_get_samples(self):
        from hydra_genetics.utils.samples import get_samples
        samples = get_samples(self.samples)
        self.assertEqual(len(samples), 5)
        self.assertTrue('NA12878' in samples)
        self.assertTrue('NA13878' in samples)
        self.assertTrue('NA22878' in samples)
        self.assertTrue('NA12978' in samples)
        self.assertTrue('BE12878' in samples)
